# -*- coding: utf-8 -*-
#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from __future__ import unicode_literals

import asyncio
import json
import re
import warnings

import pytest
from mock import patch

from elasticsearch import AsyncTransport
from elasticsearch.connection import Connection
from elasticsearch.connection_pool import DummyConnectionPool
from elasticsearch.exceptions import (
    AuthenticationException,
    AuthorizationException,
    ConnectionError,
    ElasticsearchWarning,
    NotFoundError,
    TransportError,
    UnsupportedProductError,
)
from elasticsearch.transport import _ProductChecker

pytestmark = pytest.mark.asyncio


class DummyConnection(Connection):
    def __init__(self, **kwargs):
        self.exception = kwargs.pop("exception", None)
        self.status, self.data = kwargs.pop("status", 200), kwargs.pop("data", "{}")
        self.headers = kwargs.pop("headers", {})
        self.delay = kwargs.pop("delay", 0)
        self.calls = []
        self.closed = False
        super(DummyConnection, self).__init__(**kwargs)

    async def perform_request(self, *args, **kwargs):
        if self.closed:
            raise RuntimeError("This connection is closed")
        if self.delay:
            await asyncio.sleep(self.delay)
        self.calls.append((args, kwargs))
        if self.exception:
            raise self.exception
        return self.status, self.headers, self.data

    async def close(self):
        if self.closed:
            raise RuntimeError("This connection is already closed")
        self.closed = True


CLUSTER_NODES = """{
  "_nodes" : {
    "total" : 1,
    "successful" : 1,
    "failed" : 0
  },
  "cluster_name" : "elasticsearch",
  "nodes" : {
    "SRZpKFZdQguhhvifmN6UVA" : {
      "name" : "SRZpKFZ",
      "transport_address" : "127.0.0.1:9300",
      "host" : "127.0.0.1",
      "ip" : "127.0.0.1",
      "version" : "5.0.0",
      "build_hash" : "253032b",
      "roles" : [ "master", "data", "ingest" ],
      "http" : {
        "bound_address" : [ "[fe80::1]:9200", "[::1]:9200", "127.0.0.1:9200" ],
        "publish_address" : "1.1.1.1:123",
        "max_content_length_in_bytes" : 104857600
      }
    }
  }
}"""

CLUSTER_NODES_7x_PUBLISH_HOST = """{
  "_nodes" : {
    "total" : 1,
    "successful" : 1,
    "failed" : 0
  },
  "cluster_name" : "elasticsearch",
  "nodes" : {
    "SRZpKFZdQguhhvifmN6UVA" : {
      "name" : "SRZpKFZ",
      "transport_address" : "127.0.0.1:9300",
      "host" : "127.0.0.1",
      "ip" : "127.0.0.1",
      "version" : "5.0.0",
      "build_hash" : "253032b",
      "roles" : [ "master", "data", "ingest" ],
      "http" : {
        "bound_address" : [ "[fe80::1]:9200", "[::1]:9200", "127.0.0.1:9200" ],
        "publish_address" : "somehost.tld/1.1.1.1:123",
        "max_content_length_in_bytes" : 104857600
      }
    }
  }
}"""


class TestTransport:
    async def test_single_connection_uses_dummy_connection_pool(self):
        t = AsyncTransport([{}])
        await t._async_call()
        assert isinstance(t.connection_pool, DummyConnectionPool)
        t = AsyncTransport([{"host": "localhost"}])
        await t._async_call()
        assert isinstance(t.connection_pool, DummyConnectionPool)

    async def test_request_timeout_extracted_from_params_and_passed(self):
        t = AsyncTransport([{}], connection_class=DummyConnection, meta_header=False)
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", params={"request_timeout": 42})
        assert 1 == len(t.get_connection().calls)
        assert ("GET", "/", {}, None) == t.get_connection().calls[0][0]
        assert {
            "timeout": 42,
            "ignore": (),
            "headers": None,
        } == t.get_connection().calls[0][1]

    async def test_opaque_id(self):
        t = AsyncTransport(
            [{}], opaque_id="app-1", connection_class=DummyConnection, meta_header=False
        )
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/")
        assert 1 == len(t.get_connection().calls)
        assert ("GET", "/", None, None) == t.get_connection().calls[0][0]
        assert {
            "timeout": None,
            "ignore": (),
            "headers": None,
        } == t.get_connection().calls[0][1]

        # Now try with an 'x-opaque-id' set on perform_request().
        await t.perform_request("GET", "/", headers={"x-opaque-id": "request-1"})
        assert 2 == len(t.get_connection().calls)
        assert ("GET", "/", None, None) == t.get_connection().calls[1][0]
        assert {
            "timeout": None,
            "ignore": (),
            "headers": {"x-opaque-id": "request-1"},
        } == t.get_connection().calls[1][1]

    async def test_request_with_custom_user_agent_header(self):
        t = AsyncTransport([{}], connection_class=DummyConnection, meta_header=False)
        t._verified_elasticsearch = True

        await t.perform_request(
            "GET", "/", headers={"user-agent": "my-custom-value/1.2.3"}
        )
        assert 1 == len(t.get_connection().calls)
        assert {
            "timeout": None,
            "ignore": (),
            "headers": {"user-agent": "my-custom-value/1.2.3"},
        } == t.get_connection().calls[0][1]

    async def test_send_get_body_as_source(self):
        with warnings.catch_warnings(record=True) as w:
            t = AsyncTransport(
                [{}], send_get_body_as="source", connection_class=DummyConnection
            )
        assert len(w) == 1
        assert str(w[0].message) == (
            "The 'send_get_body_as' parameter is no longer necessary and will be removed in 8.0"
        )
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body={})
        assert 1 == len(t.get_connection().calls)
        assert ("GET", "/", {"source": "{}"}, None) == t.get_connection().calls[0][0]

    async def test_send_get_body_as_post(self):
        with warnings.catch_warnings(record=True) as w:
            t = AsyncTransport(
                [{}], send_get_body_as="POST", connection_class=DummyConnection
            )
        assert len(w) == 1
        assert str(w[0].message) == (
            "The 'send_get_body_as' parameter is no longer necessary and will be removed in 8.0"
        )
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body={})
        assert 1 == len(t.get_connection().calls)
        assert ("POST", "/", None, b"{}") == t.get_connection().calls[0][0]

    async def test_client_meta_header(self):
        t = AsyncTransport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body={})
        assert len(t.get_connection().calls) == 1
        headers = t.get_connection().calls[0][1]["headers"]
        assert re.match(
            r"^es=[0-9.]+p?,py=[0-9.]+p?,t=[0-9.]+p?$",
            headers["x-elastic-client-meta"],
        )

        class DummyConnectionWithMeta(DummyConnection):
            HTTP_CLIENT_META = ("dm", "1.2.3")

        t = AsyncTransport([{}], connection_class=DummyConnectionWithMeta)
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body={}, headers={"Custom": "header"})
        assert len(t.get_connection().calls) == 1
        headers = t.get_connection().calls[0][1]["headers"]
        assert re.match(
            r"^es=[0-9.]+p?,py=[0-9.]+p?,t=[0-9.]+p?,dm=1.2.3$",
            headers["x-elastic-client-meta"],
        )
        assert headers["Custom"] == "header"

    async def test_client_meta_header_not_sent(self):
        t = AsyncTransport([{}], meta_header=False, connection_class=DummyConnection)
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body={})
        assert len(t.get_connection().calls) == 1
        headers = t.get_connection().calls[0][1]["headers"]
        assert headers is None

    async def test_body_gets_encoded_into_bytes(self):
        t = AsyncTransport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body="你好")
        assert 1 == len(t.get_connection().calls)
        assert (
            "GET",
            "/",
            None,
            b"\xe4\xbd\xa0\xe5\xa5\xbd",
        ) == t.get_connection().calls[0][0]

    async def test_body_bytes_get_passed_untouched(self):
        t = AsyncTransport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        body = b"\xe4\xbd\xa0\xe5\xa5\xbd"
        await t.perform_request("GET", "/", body=body)
        assert 1 == len(t.get_connection().calls)
        assert ("GET", "/", None, body) == t.get_connection().calls[0][0]

    async def test_body_surrogates_replaced_encoded_into_bytes(self):
        t = AsyncTransport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        await t.perform_request("GET", "/", body="你好\uda6a")
        assert 1 == len(t.get_connection().calls)
        assert (
            "GET",
            "/",
            None,
            b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa",
        ) == t.get_connection().calls[0][0]

    async def test_kwargs_passed_on_to_connections(self):
        t = AsyncTransport([{"host": "google.com"}], port=123)
        t._verified_elasticsearch = True

        await t._async_call()
        assert 1 == len(t.connection_pool.connections)
        assert "http://google.com:123" == t.connection_pool.connections[0].host

    async def test_kwargs_passed_on_to_connection_pool(self):
        dt = object()
        t = AsyncTransport([{}, {}], dead_timeout=dt)
        t._verified_elasticsearch = True

        await t._async_call()
        assert dt is t.connection_pool.dead_timeout

    async def test_custom_connection_class(self):
        class MyConnection(object):
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        t = AsyncTransport([{}], connection_class=MyConnection)
        t._verified_elasticsearch = True

        await t._async_call()
        assert 1 == len(t.connection_pool.connections)
        assert isinstance(t.connection_pool.connections[0], MyConnection)

    def test_add_connection(self):
        t = AsyncTransport([{}], randomize_hosts=False)
        t._verified_elasticsearch = True
        t.add_connection({"host": "google.com", "port": 1234})

        assert 2 == len(t.connection_pool.connections)
        assert "http://google.com:1234" == t.connection_pool.connections[1].host

    async def test_request_will_fail_after_X_retries(self):
        t = AsyncTransport(
            [{"exception": ConnectionError("abandon ship")}],
            connection_class=DummyConnection,
        )
        t._verified_elasticsearch = True

        connection_error = False
        try:
            await t.perform_request("GET", "/")
        except ConnectionError:
            connection_error = True

        assert connection_error
        assert 4 == len(t.get_connection().calls)

    async def test_failed_connection_will_be_marked_as_dead(self):
        t = AsyncTransport(
            [{"exception": ConnectionError("abandon ship")}] * 2,
            connection_class=DummyConnection,
        )
        t._verified_elasticsearch = True

        connection_error = False
        try:
            await t.perform_request("GET", "/")
        except ConnectionError:
            connection_error = True

        assert connection_error
        assert 0 == len(t.connection_pool.connections)

    async def test_resurrected_connection_will_be_marked_as_live_on_success(self):
        for method in ("GET", "HEAD"):
            t = AsyncTransport([{}, {}], connection_class=DummyConnection)
            t._verified_elasticsearch = True

            await t._async_call()
            con1 = t.connection_pool.get_connection()
            con2 = t.connection_pool.get_connection()
            t.connection_pool.mark_dead(con1)
            t.connection_pool.mark_dead(con2)

            await t.perform_request(method, "/")
            assert 1 == len(t.connection_pool.connections)
            assert 1 == len(t.connection_pool.dead_count)

    async def test_sniff_will_use_seed_connections(self):
        t = AsyncTransport([{"data": CLUSTER_NODES}], connection_class=DummyConnection)
        await t._async_call()
        t.set_connections([{"data": "invalid"}])

        await t.sniff_hosts()
        assert 1 == len(t.connection_pool.connections)
        assert "http://1.1.1.1:123" == t.get_connection().host

    async def test_sniff_on_start_fetches_and_uses_nodes_list(self):
        t = AsyncTransport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_start=True,
        )

        await t._async_call()
        await t.sniffing_task  # Need to wait for the sniffing task to complete

        assert 1 == len(t.connection_pool.connections)
        assert "http://1.1.1.1:123" == t.get_connection().host

    async def test_sniff_on_start_ignores_sniff_timeout(self):
        t = AsyncTransport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_start=True,
            sniff_timeout=12,
        )

        await t._async_call()
        await t.sniffing_task  # Need to wait for the sniffing task to complete

        assert (("GET", "/_nodes/_all/http"), {"timeout": None}) == t.seed_connections[
            0
        ].calls[0]

    async def test_sniff_uses_sniff_timeout(self):
        t = AsyncTransport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_timeout=42,
        )
        await t._async_call()
        await t.sniff_hosts()

        assert (("GET", "/_nodes/_all/http"), {"timeout": 42}) == t.seed_connections[
            0
        ].calls[0]

    async def test_sniff_reuses_connection_instances_if_possible(self):
        t = AsyncTransport(
            [{"data": CLUSTER_NODES}, {"host": "1.1.1.1", "port": 123}],
            connection_class=DummyConnection,
            randomize_hosts=False,
        )
        await t._async_call()
        connection = t.connection_pool.connections[1]
        connection.delay = 3.0  # Add this delay to make the sniffing deterministic.

        await t.sniff_hosts()
        assert 1 == len(t.connection_pool.connections)
        assert connection is t.get_connection()

    async def test_sniff_on_fail_triggers_sniffing_on_fail(self):
        t = AsyncTransport(
            [{"exception": ConnectionError("abandon ship")}, {"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_connection_fail=True,
            max_retries=0,
            randomize_hosts=False,
        )
        t._verified_elasticsearch = True

        await t._async_call()

        connection_error = False
        try:
            await t.perform_request("GET", "/")
        except ConnectionError:
            connection_error = True

        await t.sniffing_task  # Need to wait for the sniffing task to complete

        assert connection_error
        assert 1 == len(t.connection_pool.connections)
        assert "http://1.1.1.1:123" == t.get_connection().host

    @patch("elasticsearch._async.transport.AsyncTransport.sniff_hosts")
    async def test_sniff_on_fail_failing_does_not_prevent_retires(self, sniff_hosts):
        sniff_hosts.side_effect = [TransportError("sniff failed")]
        t = AsyncTransport(
            [{"exception": ConnectionError("abandon ship")}, {"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_connection_fail=True,
            max_retries=3,
            randomize_hosts=False,
        )
        t._verified_elasticsearch = True
        await t._async_init()

        conn_err, conn_data = t.connection_pool.connections
        response = await t.perform_request("GET", "/")
        assert json.loads(CLUSTER_NODES) == response
        assert 1 == sniff_hosts.call_count
        assert 1 == len(conn_err.calls)
        assert 1 == len(conn_data.calls)

    async def test_sniff_after_n_seconds(self, event_loop):
        t = AsyncTransport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniffer_timeout=5,
        )
        t._verified_elasticsearch = True
        await t._async_call()

        for _ in range(4):
            await t.perform_request("GET", "/")
        assert 1 == len(t.connection_pool.connections)
        assert isinstance(t.get_connection(), DummyConnection)
        t.last_sniff = event_loop.time() - 5.1

        await t.perform_request("GET", "/")
        await t.sniffing_task  # Need to wait for the sniffing task to complete

        assert 1 == len(t.connection_pool.connections)
        assert "http://1.1.1.1:123" == t.get_connection().host
        assert event_loop.time() - 1 < t.last_sniff < event_loop.time() + 0.01

    async def test_sniff_7x_publish_host(self):
        # Test the response shaped when a 7.x node has publish_host set
        # and the returend data is shaped in the fqdn/ip:port format.
        t = AsyncTransport(
            [{"data": CLUSTER_NODES_7x_PUBLISH_HOST}],
            connection_class=DummyConnection,
            sniff_timeout=42,
        )
        t._verified_elasticsearch = True
        await t._async_call()

        await t.sniff_hosts()
        # Ensure we parsed out the fqdn and port from the fqdn/ip:port string.
        assert t.connection_pool.connection_opts[0][1] == {
            "host": "somehost.tld",
            "port": 123,
        }

    @patch("elasticsearch._async.transport.AsyncTransport.sniff_hosts")
    async def test_sniffing_disabled_on_cloud_instances(self, sniff_hosts):
        t = AsyncTransport(
            [{}],
            sniff_on_start=True,
            sniff_on_connection_fail=True,
            connection_class=DummyConnection,
            cloud_id="cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5NyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5Ng==",
        )
        t._verified_elasticsearch = True
        await t._async_call()

        assert not t.sniff_on_connection_fail
        assert sniff_hosts.call_args is None  # Assert not called.
        await t.perform_request("GET", "/", body={})
        assert 1 == len(t.get_connection().calls)
        assert ("GET", "/", None, b"{}") == t.get_connection().calls[0][0]

    async def test_transport_close_closes_all_pool_connections(self):
        t = AsyncTransport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True
        await t._async_call()

        assert not any([conn.closed for conn in t.connection_pool.connections])
        await t.close()
        assert all([conn.closed for conn in t.connection_pool.connections])

        t = AsyncTransport([{}, {}], connection_class=DummyConnection)
        t._verified_elasticsearch = True
        await t._async_call()

        assert not any([conn.closed for conn in t.connection_pool.connections])
        await t.close()
        assert all([conn.closed for conn in t.connection_pool.connections])

    async def test_sniff_on_start_error_if_no_sniffed_hosts(self, event_loop):
        t = AsyncTransport(
            [
                {"data": ""},
                {"data": ""},
                {"data": ""},
            ],
            connection_class=DummyConnection,
            sniff_on_start=True,
        )
        t._verified_elasticsearch = True

        # If our initial sniffing attempt comes back
        # empty then we raise an error.
        with pytest.raises(TransportError) as e:
            await t._async_call()
        assert str(e.value) == "TransportError(N/A, 'Unable to sniff hosts.')"

    async def test_sniff_on_start_waits_for_sniff_to_complete(self, event_loop):
        t = AsyncTransport(
            [
                {"delay": 1, "data": ""},
                {"delay": 1, "data": ""},
                {"delay": 1, "data": CLUSTER_NODES},
            ],
            connection_class=DummyConnection,
            sniff_on_start=True,
        )
        t._verified_elasticsearch = True

        # Start the timer right before the first task
        # and have a bunch of tasks come in immediately.
        tasks = []
        start_time = event_loop.time()
        for _ in range(5):
            tasks.append(event_loop.create_task(t._async_call()))
            await asyncio.sleep(0)  # Yield to the loop

        assert t.sniffing_task is not None

        # Tasks streaming in later.
        for _ in range(5):
            tasks.append(event_loop.create_task(t._async_call()))
            await asyncio.sleep(0.1)

        # Now that all the API calls have come in we wait for
        # them all to resolve before
        await asyncio.gather(*tasks)
        end_time = event_loop.time()
        duration = end_time - start_time

        # All the tasks blocked on the sniff of each node
        # and then resolved immediately after.
        assert 1 <= duration < 2

    async def test_sniff_on_start_close_unlocks_async_calls(self, event_loop):
        t = AsyncTransport(
            [
                {"delay": 10, "data": CLUSTER_NODES},
            ],
            connection_class=DummyConnection,
            sniff_on_start=True,
        )
        t._verified_elasticsearch = True

        # Start making _async_calls() before we cancel
        tasks = []
        start_time = event_loop.time()
        for _ in range(3):
            tasks.append(event_loop.create_task(t._async_call()))
            await asyncio.sleep(0)

        # Close the transport while the sniffing task is active! :(
        await t.close()

        # Now we start waiting on all those _async_calls()
        await asyncio.gather(*tasks)
        end_time = event_loop.time()
        duration = end_time - start_time

        # A lot quicker than 10 seconds defined in 'delay'
        assert duration < 1

    @pytest.mark.parametrize(
        ["headers", "data"],
        [
            (
                {},
                '{"version":{"number":"6.99.0"},"tagline":"You Know, for Search"}',
            ),
            (
                {},
                '{"version":{"number":"7.13.0","build_flavor":"default"},"tagline":"You Know, for Search"}',
            ),
            (
                {"X-elastic-product": "Elasticsearch"},
                '{"version":{"number":"7.14.0","build_flavor":"default"},"tagline":"You Know, for Search"}',
            ),
        ],
    )
    async def test_verify_elasticsearch(self, headers, data):
        t = AsyncTransport(
            [{"data": data, "headers": headers}], connection_class=DummyConnection
        )
        await t.perform_request("GET", "/_search")
        assert t._verified_elasticsearch is True

        calls = t.connection_pool.connections[0].calls
        _ = [call[1]["headers"].pop("x-elastic-client-meta") for call in calls]

        assert calls == [
            (
                ("GET", "/"),
                {
                    "headers": {
                        "accept": "application/json",
                    },
                    "timeout": None,
                },
            ),
            (
                ("GET", "/_search", None, None),
                {
                    "headers": {},
                    "ignore": (),
                    "timeout": None,
                },
            ),
        ]

    @pytest.mark.parametrize(
        "exception_cls", [AuthorizationException, AuthenticationException]
    )
    async def test_verify_elasticsearch_skips_on_auth_errors(self, exception_cls):
        t = AsyncTransport(
            [{"exception": exception_cls(exception_cls.status_code)}],
            connection_class=DummyConnection,
        )

        with pytest.warns(ElasticsearchWarning) as warns:
            with pytest.raises(exception_cls):
                await t.perform_request(
                    "GET", "/_search", headers={"Authorization": "testme"}
                )

        # Assert that a warning was raised due to security privileges
        assert [str(w.message) for w in warns] == [
            "The client is unable to verify that the server is "
            "Elasticsearch due security privileges on the server side"
        ]

        # Assert that the cluster is "verified"
        assert t._verified_elasticsearch is True

        # See that the headers were passed along to the "info" request made
        calls = t.connection_pool.connections[0].calls
        _ = [call[1]["headers"].pop("x-elastic-client-meta") for call in calls]

        assert calls == [
            (
                ("GET", "/"),
                {
                    "headers": {
                        "accept": "application/json",
                        "authorization": "testme",
                    },
                    "timeout": None,
                },
            ),
            (
                ("GET", "/_search", None, None),
                {
                    "headers": {
                        "Authorization": "testme",
                    },
                    "ignore": (),
                    "timeout": None,
                },
            ),
        ]

    async def test_multiple_requests_verify_elasticsearch_success(self, event_loop):
        t = AsyncTransport(
            [
                {
                    "data": '{"version":{"number":"7.13.0","build_flavor":"default"},"tagline":"You Know, for Search"}',
                    "delay": 1,
                }
            ],
            connection_class=DummyConnection,
        )

        results = []
        completed_at = []

        async def request_task():
            try:
                results.append(await t.perform_request("GET", "/_search"))
            except Exception as e:
                results.append(e)
            completed_at.append(event_loop.time())

        # Execute a bunch of requests concurrently.
        tasks = []
        start_time = event_loop.time()
        for _ in range(10):
            tasks.append(event_loop.create_task(request_task()))
        await asyncio.gather(*tasks)
        end_time = event_loop.time()

        # Exactly 10 results completed
        assert len(results) == 10

        # No errors in the results
        assert all(isinstance(result, dict) for result in results)

        # Assert that this took longer than 2 seconds but less than 2.1 seconds
        duration = end_time - start_time
        assert 2 <= duration <= 2.1

        # Assert that every result came after ~2 seconds, no fast completions.
        assert all(
            2 <= completed_time - start_time <= 2.1 for completed_time in completed_at
        )

        # Assert that the cluster is "verified"
        assert t._verified_elasticsearch is True

        # See that the first request is always 'GET /' for ES check
        calls = t.connection_pool.connections[0].calls
        assert calls[0][0] == ("GET", "/")

        # The rest of the requests are 'GET /_search' afterwards
        assert all(call[0][:2] == ("GET", "/_search") for call in calls[1:])

    @pytest.mark.parametrize(
        ["build_flavor", "tagline", "product_error", "error_message"],
        [
            (
                "default",
                "BAD TAGLINE",
                _ProductChecker.UNSUPPORTED_PRODUCT,
                "The client noticed that the server is not Elasticsearch and we do not support this unknown product",
            ),
            (
                "BAD BUILD FLAVOR",
                "BAD TAGLINE",
                _ProductChecker.UNSUPPORTED_PRODUCT,
                "The client noticed that the server is not Elasticsearch and we do not support this unknown product",
            ),
            (
                "BAD BUILD FLAVOR",
                "You Know, for Search",
                _ProductChecker.UNSUPPORTED_DISTRIBUTION,
                "The client noticed that the server is not a supported distribution of Elasticsearch",
            ),
        ],
    )
    async def test_multiple_requests_verify_elasticsearch_product_error(
        self, event_loop, build_flavor, tagline, product_error, error_message
    ):
        t = AsyncTransport(
            [
                {
                    "data": '{"version":{"number":"7.13.0","build_flavor":"%s"},"tagline":"%s"}'
                    % (build_flavor, tagline),
                    "delay": 1,
                }
            ],
            connection_class=DummyConnection,
        )

        results = []
        completed_at = []

        async def request_task():
            try:
                results.append(await t.perform_request("GET", "/_search"))
            except Exception as e:
                results.append(e)
            completed_at.append(event_loop.time())

        # Execute a bunch of requests concurrently.
        tasks = []
        start_time = event_loop.time()
        for _ in range(10):
            tasks.append(event_loop.create_task(request_task()))
        await asyncio.gather(*tasks)
        end_time = event_loop.time()

        # Exactly 10 results completed
        assert len(results) == 10

        # All results were errors
        assert all(isinstance(result, UnsupportedProductError) for result in results)
        assert all(str(result) == error_message for result in results)

        # Assert that one request was made but not 2 requests.
        duration = end_time - start_time
        assert 1 <= duration <= 1.1

        # Assert that every result came after ~1 seconds, no fast completions.
        assert all(
            1 <= completed_time - start_time <= 1.1 for completed_time in completed_at
        )

        # Assert that the cluster is definitely not Elasticsearch
        assert t._verified_elasticsearch == product_error

        # See that the first request is always 'GET /' for ES check
        calls = t.connection_pool.connections[0].calls
        assert calls[0][0] == ("GET", "/")

        # The rest of the requests are 'GET /_search' afterwards
        assert all(call[0][:2] == ("GET", "/_search") for call in calls[1:])

    @pytest.mark.parametrize("error_cls", [ConnectionError, NotFoundError])
    async def test_multiple_requests_verify_elasticsearch_retry_on_errors(
        self, event_loop, error_cls
    ):
        t = AsyncTransport(
            [
                {
                    "exception": error_cls(),
                    "delay": 0.1,
                }
            ],
            connection_class=DummyConnection,
        )

        results = []
        completed_at = []

        async def request_task():
            try:
                results.append(await t.perform_request("GET", "/_search"))
            except Exception as e:
                results.append(e)
            completed_at.append(event_loop.time())

        # Execute a bunch of requests concurrently.
        tasks = []
        start_time = event_loop.time()
        for _ in range(5):
            tasks.append(event_loop.create_task(request_task()))
        await asyncio.gather(*tasks)
        end_time = event_loop.time()

        # Exactly 5 results completed
        assert len(results) == 5

        # All results were errors and not wrapped in 'NotElasticsearchError'
        assert all(isinstance(result, error_cls) for result in results)

        # Assert that 5 requests were made in total (5 transport requests per x 0.1s/conn request)
        duration = end_time - start_time
        assert 0.5 <= duration <= 0.6

        # Assert that the cluster is still in the unknown/unverified stage.
        assert t._verified_elasticsearch is None

        # See that the API isn't hit, instead it's the index requests that are failing.
        calls = t.connection_pool.connections[0].calls
        assert len(calls) == 5
        assert all(call[0] == ("GET", "/") for call in calls)



# -*- coding: utf-8 -*-
#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from __future__ import unicode_literals

from . import ElasticsearchTestCase


class TestUnicode(ElasticsearchTestCase):
    def test_indices_analyze(self):
        self.client.indices.analyze(body='{"text": "привет"}')


class TestBulk(ElasticsearchTestCase):
    def test_bulk_works_with_string_body(self):
        docs = '{ "index" : { "_index" : "bulk_test_index", "_id" : "1" } }\n{"answer": 42}'
        response = self.client.bulk(body=docs)

        self.assertFalse(response["errors"])
        self.assertEqual(1, len(response["items"]))

    def test_bulk_works_with_bytestring_body(self):
        docs = b'{ "index" : { "_index" : "bulk_test_index", "_id" : "2" } }\n{"answer": 42}'
        response = self.client.bulk(body=docs)

        self.assertFalse(response["errors"])
        self.assertEqual(1, len(response["items"]))



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from .utils import NamespacedClient, _make_path, query_params


class MigrationClient(NamespacedClient):
    @query_params(
        response_mimetypes=["application/json"],
    )
    def deprecations(self, index=None, params=None, headers=None):
        """
        Retrieves information about different cluster, node, and index level settings
        that use deprecated features that will be removed or changed in the next major
        version.

        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.16/migration-api-deprecation.html>`_

        :arg index: Index pattern
        """
        return self.transport.perform_request(
            "GET",
            _make_path(index, "_migration", "deprecations"),
            params=params,
            headers=headers,
        )

    @query_params(
        response_mimetypes=["application/json"],
    )
    def get_feature_upgrade_status(self, params=None, headers=None):
        """
        Find out whether system features need to be upgraded or not

        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.16/migration-api-feature-upgrade.html>`_
        """
        return self.transport.perform_request(
            "GET", "/_migration/system_features", params=params, headers=headers
        )

    @query_params(
        response_mimetypes=["application/json"],
    )
    def post_feature_upgrade(self, params=None, headers=None):
        """
        Begin upgrades for system features

        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.16/migration-api-feature-upgrade.html>`_
        """
        return self.transport.perform_request(
            "POST", "/_migration/system_features", params=params, headers=headers
        )



# -*- coding: utf-8 -*-
#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import threading
import time

import mock
import pytest

from elasticsearch import Elasticsearch, helpers
from elasticsearch.helpers import actions
from elasticsearch.serializer import JSONSerializer

from .test_cases import TestCase

lock_side_effect = threading.Lock()


def mock_process_bulk_chunk(*args, **kwargs):
    """
    Threadsafe way of mocking process bulk chunk:
    https://stackoverflow.com/questions/39332139/thread-safe-version-of-mock-call-count
    """

    with lock_side_effect:
        mock_process_bulk_chunk.call_count += 1
    time.sleep(0.1)
    return []


mock_process_bulk_chunk.call_count = 0


class TestParallelBulk(TestCase):
    @mock.patch(
        "elasticsearch.helpers.actions._process_bulk_chunk",
        side_effect=mock_process_bulk_chunk,
    )
    def test_all_chunks_sent(self, _process_bulk_chunk):
        actions = ({"x": i} for i in range(100))
        list(helpers.parallel_bulk(Elasticsearch(), actions, chunk_size=2))

        self.assertEqual(50, mock_process_bulk_chunk.call_count)

    @pytest.mark.skip
    @mock.patch(
        "elasticsearch.helpers.actions._process_bulk_chunk",
        # make sure we spend some time in the thread
        side_effect=lambda *a: [
            (True, time.sleep(0.001) or threading.current_thread().ident)
        ],
    )
    def test_chunk_sent_from_different_threads(self, _process_bulk_chunk):
        actions = ({"x": i} for i in range(100))
        results = list(
            helpers.parallel_bulk(
                Elasticsearch(), actions, thread_count=10, chunk_size=2
            )
        )
        self.assertTrue(len(set([r[1] for r in results])) > 1)


class TestChunkActions(TestCase):
    def setup_method(self, _):
        self.actions = [({"index": {}}, {"some": u"datá", "i": i}) for i in range(100)]

    def test_expand_action(self):
        self.assertEqual(helpers.expand_action({}), ({"index": {}}, {}))
        self.assertEqual(
            helpers.expand_action({"key": "val"}), ({"index": {}}, {"key": "val"})
        )

    def test_expand_action_actions(self):
        self.assertEqual(
            helpers.expand_action(
                {"_op_type": "delete", "_id": "id", "_index": "index"}
            ),
            ({"delete": {"_id": "id", "_index": "index"}}, None),
        )
        self.assertEqual(
            helpers.expand_action(
                {"_op_type": "update", "_id": "id", "_index": "index", "key": "val"}
            ),
            ({"update": {"_id": "id", "_index": "index"}}, {"key": "val"}),
        )
        self.assertEqual(
            helpers.expand_action(
                {"_op_type": "create", "_id": "id", "_index": "index", "key": "val"}
            ),
            ({"create": {"_id": "id", "_index": "index"}}, {"key": "val"}),
        )
        self.assertEqual(
            helpers.expand_action(
                {
                    "_op_type": "create",
                    "_id": "id",
                    "_index": "index",
                    "_source": {"key": "val"},
                }
            ),
            ({"create": {"_id": "id", "_index": "index"}}, {"key": "val"}),
        )

    def test_expand_action_options(self):
        for option in (
            "_id",
            "_index",
            "_percolate",
            "_timestamp",
            "_type",
            "if_seq_no",
            "if_primary_term",
            "parent",
            "pipeline",
            "retry_on_conflict",
            "routing",
            "version",
            "version_type",
            ("_parent", "parent"),
            ("_retry_on_conflict", "retry_on_conflict"),
            ("_routing", "routing"),
            ("_version", "version"),
            ("_version_type", "version_type"),
            ("_if_seq_no", "if_seq_no"),
            ("_if_primary_term", "if_primary_term"),
        ):
            if isinstance(option, str):
                action_option = option
            else:
                option, action_option = option
            self.assertEqual(
                helpers.expand_action({"key": "val", option: 0}),
                ({"index": {action_option: 0}}, {"key": "val"}),
            )

    def test__source_metadata_or_source(self):
        self.assertEqual(
            helpers.expand_action({"_source": {"key": "val"}}),
            ({"index": {}}, {"key": "val"}),
        )

        self.assertEqual(
            helpers.expand_action(
                {"_source": ["key"], "key": "val", "_op_type": "update"}
            ),
            ({"update": {"_source": ["key"]}}, {"key": "val"}),
        )

        self.assertEqual(
            helpers.expand_action(
                {"_source": True, "key": "val", "_op_type": "update"}
            ),
            ({"update": {"_source": True}}, {"key": "val"}),
        )

        # This case is only to ensure backwards compatibility with old functionality.
        self.assertEqual(
            helpers.expand_action(
                {"_source": {"key2": "val2"}, "key": "val", "_op_type": "update"}
            ),
            ({"update": {}}, {"key2": "val2"}),
        )

    def test_chunks_are_chopped_by_byte_size(self):
        self.assertEqual(
            100,
            len(
                list(helpers._chunk_actions(self.actions, 100000, 1, JSONSerializer()))
            ),
        )

    def test_chunks_are_chopped_by_chunk_size(self):
        self.assertEqual(
            10,
            len(
                list(
                    helpers._chunk_actions(self.actions, 10, 99999999, JSONSerializer())
                )
            ),
        )

    def test_chunks_are_chopped_by_byte_size_properly(self):
        max_byte_size = 170
        chunks = list(
            helpers._chunk_actions(
                self.actions, 100000, max_byte_size, JSONSerializer()
            )
        )
        self.assertEqual(25, len(chunks))
        for chunk_data, chunk_actions in chunks:
            chunk = u"".join(chunk_actions)
            chunk = chunk if isinstance(chunk, str) else chunk.encode("utf-8")
            self.assertLessEqual(len(chunk), max_byte_size)

    def test_add_helper_meta_to_kwargs(self):
        self.assertEqual(
            actions._add_helper_meta_to_kwargs({}, "b"),
            {"params": {"__elastic_client_meta": (("h", "b"),)}},
        )
        self.assertEqual(
            actions._add_helper_meta_to_kwargs({"params": {}}, "b"),
            {"params": {"__elastic_client_meta": (("h", "b"),)}},
        )
        self.assertEqual(
            actions._add_helper_meta_to_kwargs({"params": {"key": "value"}}, "b"),
            {"params": {"__elastic_client_meta": (("h", "b"),), "key": "value"}},
        )


class TestExpandActions(TestCase):
    def test_string_actions_are_marked_as_simple_inserts(self):
        self.assertEqual(
            ('{"index":{}}', "whatever"), helpers.expand_action("whatever")
        )



# -*- coding: utf-8 -*-
#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from __future__ import unicode_literals

import json
import time
import warnings

import pytest
from mock import patch

from elasticsearch.connection import Connection
from elasticsearch.connection_pool import DummyConnectionPool
from elasticsearch.exceptions import (
    AuthenticationException,
    AuthorizationException,
    ConnectionError,
    ElasticsearchWarning,
    NotFoundError,
    TransportError,
    UnsupportedProductError,
)
from elasticsearch.transport import Transport, _ProductChecker, get_host_info

from .test_cases import TestCase


class DummyConnection(Connection):
    def __init__(self, **kwargs):
        self.exception = kwargs.pop("exception", None)
        self.status, self.data = kwargs.pop("status", 200), kwargs.pop("data", "{}")
        self.headers = kwargs.pop("headers", {})
        self.delay = kwargs.pop("delay", None)
        self.calls = []
        super(DummyConnection, self).__init__(**kwargs)

    def perform_request(self, *args, **kwargs):
        self.calls.append((args, kwargs))
        if self.delay is not None:
            time.sleep(self.delay)
        if self.exception:
            raise self.exception
        return self.status, self.headers, self.data


CLUSTER_NODES = """{
  "_nodes" : {
    "total" : 1,
    "successful" : 1,
    "failed" : 0
  },
  "cluster_name" : "elasticsearch",
  "nodes" : {
    "SRZpKFZdQguhhvifmN6UVA" : {
      "name" : "SRZpKFZ",
      "transport_address" : "127.0.0.1:9300",
      "host" : "127.0.0.1",
      "ip" : "127.0.0.1",
      "version" : "5.0.0",
      "build_hash" : "253032b",
      "roles" : [ "master", "data", "ingest" ],
      "http" : {
        "bound_address" : [ "[fe80::1]:9200", "[::1]:9200", "127.0.0.1:9200" ],
        "publish_address" : "1.1.1.1:123",
        "max_content_length_in_bytes" : 104857600
      }
    }
  }
}"""

CLUSTER_NODES_7x_PUBLISH_HOST = """{
  "_nodes" : {
    "total" : 1,
    "successful" : 1,
    "failed" : 0
  },
  "cluster_name" : "elasticsearch",
  "nodes" : {
    "SRZpKFZdQguhhvifmN6UVA" : {
      "name" : "SRZpKFZ",
      "transport_address" : "127.0.0.1:9300",
      "host" : "127.0.0.1",
      "ip" : "127.0.0.1",
      "version" : "5.0.0",
      "build_hash" : "253032b",
      "roles" : [ "master", "data", "ingest" ],
      "http" : {
        "bound_address" : [ "[fe80::1]:9200", "[::1]:9200", "127.0.0.1:9200" ],
        "publish_address" : "somehost.tld/1.1.1.1:123",
        "max_content_length_in_bytes" : 104857600
      }
    }
  }
}"""


class TestHostsInfoCallback(TestCase):
    def test_master_only_nodes_are_ignored(self):
        nodes = [
            {"roles": ["master"]},
            {"roles": ["master", "data", "ingest"]},
            {"roles": ["data", "ingest"]},
            {"roles": []},
            {},
        ]
        chosen = [
            i
            for i, node_info in enumerate(nodes)
            if get_host_info(node_info, i) is not None
        ]
        self.assertEqual([1, 2, 3, 4], chosen)


class TestTransport(TestCase):
    def test_single_connection_uses_dummy_connection_pool(self):
        t = Transport([{}])
        t._verified_elasticsearch = True
        self.assertIsInstance(t.connection_pool, DummyConnectionPool)

        t = Transport([{"host": "localhost"}])
        t._verified_elasticsearch = True
        self.assertIsInstance(t.connection_pool, DummyConnectionPool)

    def test_request_timeout_extracted_from_params_and_passed(self):
        t = Transport([{}], meta_header=False, connection_class=DummyConnection)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", params={"request_timeout": 42})
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(("GET", "/", {}, None), t.get_connection().calls[0][0])
        self.assertEqual(
            {"timeout": 42, "ignore": (), "headers": None},
            t.get_connection().calls[0][1],
        )

    def test_opaque_id(self):
        t = Transport(
            [{}], opaque_id="app-1", meta_header=False, connection_class=DummyConnection
        )
        t._verified_elasticsearch = True

        t.perform_request("GET", "/")
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(("GET", "/", None, None), t.get_connection().calls[0][0])
        self.assertEqual(
            {"timeout": None, "ignore": (), "headers": None},
            t.get_connection().calls[0][1],
        )

        # Now try with an 'x-opaque-id' set on perform_request().
        t.perform_request("GET", "/", headers={"x-opaque-id": "request-1"})
        self.assertEqual(2, len(t.get_connection().calls))
        self.assertEqual(("GET", "/", None, None), t.get_connection().calls[1][0])
        self.assertEqual(
            {"timeout": None, "ignore": (), "headers": {"x-opaque-id": "request-1"}},
            t.get_connection().calls[1][1],
        )

    def test_request_with_custom_user_agent_header(self):
        t = Transport([{}], meta_header=False, connection_class=DummyConnection)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", headers={"user-agent": "my-custom-value/1.2.3"})
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(
            {
                "timeout": None,
                "ignore": (),
                "headers": {"user-agent": "my-custom-value/1.2.3"},
            },
            t.get_connection().calls[0][1],
        )

    def test_send_get_body_as_source(self):
        with warnings.catch_warnings(record=True) as w:
            t = Transport(
                [{}], send_get_body_as="source", connection_class=DummyConnection
            )
        assert len(w) == 1
        assert str(w[0].message) == (
            "The 'send_get_body_as' parameter is no longer necessary and will be removed in 8.0"
        )
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body={})
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(
            ("GET", "/", {"source": "{}"}, None), t.get_connection().calls[0][0]
        )

    def test_send_get_body_as_post(self):
        with warnings.catch_warnings(record=True) as w:
            t = Transport(
                [{}], send_get_body_as="POST", connection_class=DummyConnection
            )
        assert len(w) == 1
        assert str(w[0].message) == (
            "The 'send_get_body_as' parameter is no longer necessary and will be removed in 8.0"
        )
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body={})
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(("POST", "/", None, b"{}"), t.get_connection().calls[0][0])

    def test_client_meta_header(self):
        t = Transport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body={})
        self.assertEqual(1, len(t.get_connection().calls))
        headers = t.get_connection().calls[0][1]["headers"]
        self.assertRegexpMatches(
            headers["x-elastic-client-meta"], r"^es=[0-9.]+p?,py=[0-9.]+p?,t=[0-9.]+p?$"
        )

        class DummyConnectionWithMeta(DummyConnection):
            HTTP_CLIENT_META = ("dm", "1.2.3")

        t = Transport([{}], connection_class=DummyConnectionWithMeta)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body={}, headers={"Custom": "header"})
        self.assertEqual(1, len(t.get_connection().calls))
        headers = t.get_connection().calls[0][1]["headers"]
        self.assertRegexpMatches(
            headers["x-elastic-client-meta"],
            r"^es=[0-9.]+p?,py=[0-9.]+p?,t=[0-9.]+p?,dm=1.2.3$",
        )
        self.assertEqual(headers["Custom"], "header")

    def test_client_meta_header_not_sent(self):
        t = Transport([{}], meta_header=False, connection_class=DummyConnection)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body={})
        self.assertEqual(1, len(t.get_connection().calls))
        headers = t.get_connection().calls[0][1]["headers"]
        self.assertIs(headers, None)

    def test_meta_header_type_error(self):
        with pytest.raises(TypeError) as e:
            Transport([{}], meta_header=1)
        assert str(e.value) == "meta_header must be of type bool"

    def test_body_gets_encoded_into_bytes(self):
        t = Transport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body="你好")
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(
            ("GET", "/", None, b"\xe4\xbd\xa0\xe5\xa5\xbd"),
            t.get_connection().calls[0][0],
        )

    def test_body_bytes_get_passed_untouched(self):
        t = Transport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        body = b"\xe4\xbd\xa0\xe5\xa5\xbd"
        t.perform_request("GET", "/", body=body)
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(("GET", "/", None, body), t.get_connection().calls[0][0])

    def test_body_surrogates_replaced_encoded_into_bytes(self):
        t = Transport([{}], connection_class=DummyConnection)
        t._verified_elasticsearch = True

        t.perform_request("GET", "/", body="你好\uda6a")
        self.assertEqual(1, len(t.get_connection().calls))
        self.assertEqual(
            ("GET", "/", None, b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"),
            t.get_connection().calls[0][0],
        )

    def test_kwargs_passed_on_to_connections(self):
        t = Transport([{"host": "google.com"}], port=123)
        t._verified_elasticsearch = True
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertEqual("http://google.com:123", t.connection_pool.connections[0].host)

    def test_kwargs_passed_on_to_connection_pool(self):
        dt = object()
        t = Transport([{}, {}], dead_timeout=dt)
        t._verified_elasticsearch = True
        self.assertIs(dt, t.connection_pool.dead_timeout)

    def test_custom_connection_class(self):
        class MyConnection(object):
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        t = Transport([{}], connection_class=MyConnection)
        t._verified_elasticsearch = True
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertIsInstance(t.connection_pool.connections[0], MyConnection)

    def test_add_connection(self):
        t = Transport([{}], randomize_hosts=False)
        t._verified_elasticsearch = True
        t.add_connection({"host": "google.com", "port": 1234})

        self.assertEqual(2, len(t.connection_pool.connections))
        self.assertEqual(
            "http://google.com:1234", t.connection_pool.connections[1].host
        )

    def test_request_will_fail_after_X_retries(self):
        t = Transport(
            [{"exception": ConnectionError("abandon ship")}],
            connection_class=DummyConnection,
        )
        t._verified_elasticsearch = True

        self.assertRaises(ConnectionError, t.perform_request, "GET", "/")
        self.assertEqual(4, len(t.get_connection().calls))

    def test_failed_connection_will_be_marked_as_dead(self):
        t = Transport(
            [{"exception": ConnectionError("abandon ship")}] * 2,
            connection_class=DummyConnection,
        )
        t._verified_elasticsearch = True

        self.assertRaises(ConnectionError, t.perform_request, "GET", "/")
        self.assertEqual(0, len(t.connection_pool.connections))

    def test_resurrected_connection_will_be_marked_as_live_on_success(self):
        for method in ("GET", "HEAD"):
            t = Transport([{}, {}], connection_class=DummyConnection)
            t._verified_elasticsearch = True
            con1 = t.connection_pool.get_connection()
            con2 = t.connection_pool.get_connection()
            t.connection_pool.mark_dead(con1)
            t.connection_pool.mark_dead(con2)

            t.perform_request(method, "/")
            self.assertEqual(1, len(t.connection_pool.connections))
            self.assertEqual(1, len(t.connection_pool.dead_count))

    def test_sniff_will_use_seed_connections(self):
        t = Transport([{"data": CLUSTER_NODES}], connection_class=DummyConnection)
        t._verified_elasticsearch = True
        t.set_connections([{"data": "invalid"}])

        t.sniff_hosts()
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertEqual("http://1.1.1.1:123", t.get_connection().host)

    def test_sniff_on_start_fetches_and_uses_nodes_list(self):
        t = Transport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_start=True,
        )
        t._verified_elasticsearch = True
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertEqual("http://1.1.1.1:123", t.get_connection().host)

    def test_sniff_on_start_ignores_sniff_timeout(self):
        t = Transport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_start=True,
            sniff_timeout=12,
        )
        t._verified_elasticsearch = True
        self.assertEqual(
            (("GET", "/_nodes/_all/http"), {"timeout": None}),
            t.seed_connections[0].calls[0],
        )

    def test_sniff_uses_sniff_timeout(self):
        t = Transport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_timeout=42,
        )
        t._verified_elasticsearch = True
        t.sniff_hosts()
        self.assertEqual(
            (("GET", "/_nodes/_all/http"), {"timeout": 42}),
            t.seed_connections[0].calls[0],
        )

    def test_sniff_reuses_connection_instances_if_possible(self):
        t = Transport(
            [{"data": CLUSTER_NODES}, {"host": "1.1.1.1", "port": 123}],
            connection_class=DummyConnection,
            randomize_hosts=False,
        )
        t._verified_elasticsearch = True
        connection = t.connection_pool.connections[1]

        t.sniff_hosts()
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertIs(connection, t.get_connection())

    def test_sniff_on_fail_triggers_sniffing_on_fail(self):
        t = Transport(
            [{"exception": ConnectionError("abandon ship")}, {"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_connection_fail=True,
            max_retries=0,
            randomize_hosts=False,
        )
        t._verified_elasticsearch = True

        self.assertRaises(ConnectionError, t.perform_request, "GET", "/")
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertEqual("http://1.1.1.1:123", t.get_connection().host)

    @patch("elasticsearch.transport.Transport.sniff_hosts")
    def test_sniff_on_fail_failing_does_not_prevent_retires(self, sniff_hosts):
        sniff_hosts.side_effect = [TransportError("sniff failed")]
        t = Transport(
            [{"exception": ConnectionError("abandon ship")}, {"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniff_on_connection_fail=True,
            max_retries=3,
            randomize_hosts=False,
        )
        t._verified_elasticsearch = True

        conn_err, conn_data = t.connection_pool.connections
        response = t.perform_request("GET", "/")
        self.assertEqual(json.loads(CLUSTER_NODES), response)
        self.assertEqual(1, sniff_hosts.call_count)
        self.assertEqual(1, len(conn_err.calls))
        self.assertEqual(1, len(conn_data.calls))

    def test_sniff_after_n_seconds(self):
        t = Transport(
            [{"data": CLUSTER_NODES}],
            connection_class=DummyConnection,
            sniffer_timeout=5,
        )
        t._verified_elasticsearch = True

        for _ in range(4):
            t.perform_request("GET", "/")
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertIsInstance(t.get_connection(), DummyConnection)
        t.last_sniff = time.time() - 5.1

        t.perform_request("GET", "/")
        self.assertEqual(1, len(t.connection_pool.connections))
        self.assertEqual("http://1.1.1.1:123", t.get_connection().host)
        self.assertTrue(time.time() - 1 < t.last_sniff < time.time() + 0.01)

    def test_sniff_7x_publish_host(self):
        # Test the response shaped when a 7.x node has publish_host set
        # and the returend data is shaped in the fqdn/ip:port format.
        t = Transport(
            [{"data": CLUSTER_NODES_7x_PUBLISH_HOST}],
            connection_class=DummyConnection,
            sniff_timeout=42,
        )
        t._verified_elasticsearch = True
        t.sniff_hosts()
        # Ensure we parsed out the fqdn and port from the fqdn/ip:port string.
        self.assertEqual(
            t.connection_pool.connection_opts[0][1],
            {"host": "somehost.tld", "port": 123},
        )

    @patch("elasticsearch.transport.Transport.sniff_hosts")
    def test_sniffing_disabled_on_cloud_instances(self, sniff_hosts):
        t = Transport(
            [{}],
            sniff_on_start=True,
            sniff_on_connection_fail=True,
            cloud_id="cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5NyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5Ng==",
        )
        t._verified_elasticsearch = True

        self.assertFalse(t.sniff_on_connection_fail)
        self.assertIs(sniff_hosts.call_args, None)  # Assert not called.


TAGLINE = "You Know, for Search"


@pytest.mark.parametrize(
    ["headers", "response", "product_error"],
    [
        # All empty.
        ({}, {}, _ProductChecker.UNSUPPORTED_PRODUCT),
        # Don't check the product header immediately, need to check version first.
        (
            {"x-elastic-product": "Elasticsearch"},
            {},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Version not there.
        ({}, {"tagline": TAGLINE}, _ProductChecker.UNSUPPORTED_PRODUCT),
        # Version is nonsense
        (
            {},
            {"version": "1.0.0", "tagline": TAGLINE},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Version number not there
        ({}, {"version": {}, "tagline": TAGLINE}, _ProductChecker.UNSUPPORTED_PRODUCT),
        # Version number is nonsense
        (
            {},
            {"version": {"number": "nonsense"}, "tagline": TAGLINE},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Version number way in the past
        (
            {},
            {"version": {"number": "1.0.0"}, "tagline": TAGLINE},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Version number way in the future
        (
            {},
            {"version": {"number": "999.0.0"}, "tagline": TAGLINE},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Build flavor not supposed to be missing
        (
            {},
            {"version": {"number": "7.13.0"}, "tagline": TAGLINE},
            _ProductChecker.UNSUPPORTED_DISTRIBUTION,
        ),
        # Build flavor is 'oss'
        (
            {},
            {
                "version": {"number": "7.10.0", "build_flavor": "oss"},
                "tagline": TAGLINE,
            },
            _ProductChecker.UNSUPPORTED_DISTRIBUTION,
        ),
        # Build flavor is nonsense
        (
            {},
            {
                "version": {"number": "7.13.0", "build_flavor": "nonsense"},
                "tagline": TAGLINE,
            },
            _ProductChecker.UNSUPPORTED_DISTRIBUTION,
        ),
        # Tagline is nonsense
        (
            {},
            {"version": {"number": "7.1.0-SNAPSHOT"}, "tagline": "nonsense"},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Product header is not supposed to be missing
        (
            {},
            {"version": {"number": "7.14.0"}, "tagline": "You Know, for Search"},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
        # Product header is nonsense
        (
            {"x-elastic-product": "nonsense"},
            {"version": {"number": "7.15.0"}, "tagline": TAGLINE},
            _ProductChecker.UNSUPPORTED_PRODUCT,
        ),
    ],
)
def test_verify_elasticsearch_errors(headers, response, product_error):
    assert _ProductChecker.check_product(headers, response) == product_error


@pytest.mark.parametrize(
    ["headers", "response"],
    [
        ({}, {"version": {"number": "6.0.0"}, "tagline": TAGLINE}),
        ({}, {"version": {"number": "6.99.99"}, "tagline": TAGLINE}),
        (
            {},
            {
                "version": {"number": "7.0.0", "build_flavor": "default"},
                "tagline": TAGLINE,
            },
        ),
        (
            {},
            {
                "version": {"number": "7.13.99", "build_flavor": "default"},
                "tagline": TAGLINE,
            },
        ),
        (
            {"x-elastic-product": "Elasticsearch"},
            {
                "version": {"number": "7.14.0", "build_flavor": "default"},
                "tagline": TAGLINE,
            },
        ),
        (
            {"x-elastic-product": "Elasticsearch"},
            {
                "version": {"number": "7.99.99", "build_flavor": "default"},
                "tagline": TAGLINE,
            },
        ),
        (
            {"x-elastic-product": "Elasticsearch"},
            {
                "version": {"number": "8.0.0"},
            },
        ),
    ],
)
def test_verify_elasticsearch_passes(headers, response):
    result = _ProductChecker.check_product(headers, response)
    assert result == _ProductChecker.SUCCESS
    assert result is True


@pytest.mark.parametrize(
    ["headers", "data"],
    [
        (
            {},
            '{"version":{"number":"6.99.0"},"tagline":"You Know, for Search"}',
        ),
        (
            {},
            """{
  "name" : "io",
  "cluster_name" : "elasticsearch",
  "cluster_uuid" : "HaMHUswUSGGnzla8B17Iqw",
  "version" : {
    "number" : "7.6.0",
    "build_flavor" : "default",
    "build_type" : "tar",
    "build_hash" : "7f634e9f44834fbc12724506cc1da681b0c3b1e3",
    "build_date" : "2020-02-06T00:09:00.449973Z",
    "build_snapshot" : false,
    "lucene_version" : "8.4.0",
    "minimum_wire_compatibility_version" : "6.8.0",
    "minimum_index_compatibility_version" : "6.0.0-beta1"
  },
  "tagline" : "You Know, for Search"
}""",
        ),
        (
            {},
            '{"version":{"number":"7.13.0","build_flavor":"default"},"tagline":"You Know, for Search"}',
        ),
        (
            {"X-elastic-product": "Elasticsearch"},
            '{"version":{"number":"7.14.0","build_flavor":"default"},"tagline":"You Know, for Search"}',
        ),
    ],
)
def test_verify_elasticsearch(headers, data):
    t = Transport(
        [{"data": data, "headers": headers}], connection_class=DummyConnection
    )
    t.perform_request("GET", "/_search")
    assert t._verified_elasticsearch is True

    calls = t.connection_pool.connections[0].calls
    _ = [call[1]["headers"].pop("x-elastic-client-meta") for call in calls]

    assert calls == [
        (
            ("GET", "/"),
            {
                "headers": {
                    "accept": "application/json",
                },
                "timeout": None,
            },
        ),
        (
            ("GET", "/_search", None, None),
            {
                "headers": {},
                "ignore": (),
                "timeout": None,
            },
        ),
    ]


@pytest.mark.parametrize(
    "exception_cls", [AuthorizationException, AuthenticationException]
)
def test_verify_elasticsearch_skips_on_auth_errors(exception_cls):
    t = Transport(
        [{"exception": exception_cls(exception_cls.status_code)}],
        connection_class=DummyConnection,
    )

    with pytest.warns(ElasticsearchWarning) as warns:
        with pytest.raises(exception_cls):
            t.perform_request(
                "GET",
                "/_search",
                headers={"Authorization": "testme"},
                params={"request_timeout": 3},
            )

    # Assert that a warning was raised due to security privileges
    assert [str(w.message) for w in warns] == [
        "The client is unable to verify that the server is "
        "Elasticsearch due security privileges on the server side"
    ]

    # Assert that the cluster is "verified"
    assert t._verified_elasticsearch is True

    # See that the headers were passed along to the "info" request made
    calls = t.connection_pool.connections[0].calls
    _ = [call[1]["headers"].pop("x-elastic-client-meta") for call in calls]

    assert calls == [
        (
            ("GET", "/"),
            {
                "headers": {
                    "accept": "application/json",
                    "authorization": "testme",
                },
                "timeout": 3,
            },
        ),
        (
            ("GET", "/_search", {}, None),
            {
                "headers": {
                    "Authorization": "testme",
                },
                "ignore": (),
                "timeout": 3,
            },
        ),
    ]


def test_multiple_requests_verify_elasticsearch_success():
    try:
        import threading
    except ImportError:
        return pytest.skip("Requires the 'threading' module")

    t = Transport(
        [
            {
                "data": '{"version":{"number":"7.13.0","build_flavor":"default"},"tagline":"You Know, for Search"}',
                "delay": 1,
            }
        ],
        connection_class=DummyConnection,
    )

    results = []
    completed_at = []

    class RequestThread(threading.Thread):
        def run(self):
            try:
                results.append(t.perform_request("GET", "/_search"))
            except Exception as e:
                results.append(e)
            completed_at.append(time.time())

    # Execute a bunch of requests concurrently.
    threads = []
    start_time = time.time()
    for _ in range(10):
        thread = RequestThread()
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
    end_time = time.time()

    # Exactly 10 results completed
    assert len(results) == 10

    # No errors in the results
    assert all(isinstance(result, dict) for result in results)

    # Assert that this took longer than 2 seconds but less than 2.1 seconds
    duration = end_time - start_time
    assert 2 <= duration <= 2.1

    # Assert that every result came after ~2 seconds, no fast completions.
    assert all(
        2 <= completed_time - start_time <= 2.1 for completed_time in completed_at
    )

    # Assert that the cluster is "verified"
    assert t._verified_elasticsearch is True

    # See that the first request is always 'GET /' for ES check
    calls = t.connection_pool.connections[0].calls
    assert calls[0][0] == ("GET", "/")

    # The rest of the requests are 'GET /_search' afterwards
    assert all(call[0][:2] == ("GET", "/_search") for call in calls[1:])


@pytest.mark.parametrize(
    ["build_flavor", "tagline", "product_error", "error_message"],
    [
        (
            "default",
            "BAD TAGLINE",
            _ProductChecker.UNSUPPORTED_PRODUCT,
            "The client noticed that the server is not Elasticsearch and we do not support this unknown product",
        ),
        (
            "BAD BUILD FLAVOR",
            "BAD TAGLINE",
            _ProductChecker.UNSUPPORTED_PRODUCT,
            "The client noticed that the server is not Elasticsearch and we do not support this unknown product",
        ),
        (
            "BAD BUILD FLAVOR",
            "You Know, for Search",
            _ProductChecker.UNSUPPORTED_DISTRIBUTION,
            "The client noticed that the server is not a supported distribution of Elasticsearch",
        ),
    ],
)
def test_multiple_requests_verify_elasticsearch_product_error(
    build_flavor, tagline, product_error, error_message
):
    try:
        import threading
    except ImportError:
        return pytest.skip("Requires the 'threading' module")

    t = Transport(
        [
            {
                "data": '{"version":{"number":"7.13.0","build_flavor":"%s"},"tagline":"%s"}'
                % (build_flavor, tagline),
                "delay": 1,
            }
        ],
        connection_class=DummyConnection,
    )

    results = []
    completed_at = []

    class RequestThread(threading.Thread):
        def run(self):
            try:
                results.append(t.perform_request("GET", "/_search"))
            except Exception as e:
                results.append(e)
            completed_at.append(time.time())

    # Execute a bunch of requests concurrently.
    threads = []
    start_time = time.time()
    for _ in range(10):
        thread = RequestThread()
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
    end_time = time.time()

    # Exactly 10 results completed
    assert len(results) == 10

    # All results were errors
    assert all(isinstance(result, UnsupportedProductError) for result in results)
    assert all(str(result) == error_message for result in results)

    # Assert that one request was made but not 2 requests.
    duration = end_time - start_time
    assert 1 <= duration <= 1.1

    # Assert that every result came after ~1 seconds, no fast completions.
    assert all(
        1 <= completed_time - start_time <= 1.1 for completed_time in completed_at
    )

    # Assert that the cluster is definitely not Elasticsearch
    assert t._verified_elasticsearch == product_error

    # See that the first request is always 'GET /' for ES check
    calls = t.connection_pool.connections[0].calls
    assert calls[0][0] == ("GET", "/")

    # The rest of the requests are 'GET /_search' afterwards
    assert all(call[0][:2] == ("GET", "/_search") for call in calls[1:])


@pytest.mark.parametrize("error_cls", [ConnectionError, NotFoundError])
def test_multiple_requests_verify_elasticsearch_retry_on_errors(error_cls):
    try:
        import threading
    except ImportError:
        return pytest.skip("Requires the 'threading' module")

    t = Transport(
        [
            {
                "exception": error_cls(),
                "delay": 0.1,
            }
        ],
        connection_class=DummyConnection,
    )

    results = []
    completed_at = []

    class RequestThread(threading.Thread):
        def run(self):
            try:
                results.append(t.perform_request("GET", "/_search"))
            except Exception as e:
                results.append(e)
            completed_at.append(time.time())

    # Execute a bunch of requests concurrently.
    threads = []
    start_time = time.time()
    for _ in range(5):
        thread = RequestThread()
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
    end_time = time.time()

    # Exactly 5 results completed
    assert len(results) == 5

    # All results were errors and not wrapped in 'UnsupportedProductError'
    assert all(isinstance(result, error_cls) for result in results)

    # Assert that 5 requests were made in total (5 transport requests per x 0.1s/conn request)
    duration = end_time - start_time
    assert 0.5 <= duration <= 0.6

    # Assert that the cluster is still in the unknown/unverified stage.
    assert t._verified_elasticsearch is None

    # See that the API isn't hit, instead it's the index requests that are failing.
    calls = t.connection_pool.connections[0].calls
    assert len(calls) == 5
    assert all(call[0] == ("GET", "/") for call in calls)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import importlib
import sys
import warnings

import pytest

import elasticsearch


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Requires Python 3.6+")
def test_no_deprecation_python3_6_and_later():
    with warnings.catch_warnings(record=True) as w:
        importlib.reload(elasticsearch)

    assert len(w) == 0


@pytest.mark.skipif(sys.version_info >= (3, 6), reason="Requires Python <3.6")
def test_deprecated_python3_5_and_earlier():

    try:  # Python 3.4+
        import imp

        reload = imp.reload
    except ImportError:  # Python 2.7
        reload = reload

    with pytest.warns(DeprecationWarning) as w:
        reload(elasticsearch)

    assert len(w) == 1
    assert str(w[0].message) == (
        "Support for Python 3.5 and earlier is deprecated and will be removed "
        "in v8.0.0 (current instance is Python %d.%d) See https://github.com/"
        "elastic/elasticsearch-py/issues/1696 for details." % (sys.version_info[:2])
    )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

"""Script which verifies that all source files have a license header.
Has two modes: 'fix' and 'check'. 'fix' fixes problems, 'check' will
error out if 'fix' would have changed the file.
"""

import os
import re
import sys
from itertools import chain
from typing import Iterator, List

lines_to_keep = ["# -*- coding: utf-8 -*-\n", "#!/usr/bin/env python\n"]
license_header_lines = [
    "#  Licensed to Elasticsearch B.V. under one or more contributor\n",
    "#  license agreements. See the NOTICE file distributed with\n",
    "#  this work for additional information regarding copyright\n",
    "#  ownership. Elasticsearch B.V. licenses this file to you under\n",
    '#  the Apache License, Version 2.0 (the "License"); you may\n',
    "#  not use this file except in compliance with the License.\n",
    "#  You may obtain a copy of the License at\n",
    "#\n",
    "# 	http://www.apache.org/licenses/LICENSE-2.0\n",
    "#\n",
    "#  Unless required by applicable law or agreed to in writing,\n",
    "#  software distributed under the License is distributed on an\n",
    '#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\n',
    "#  KIND, either express or implied.  See the License for the\n",
    "#  specific language governing permissions and limitations\n",
    "#  under the License.\n",
    "\n",
]


def find_files_to_fix(sources: List[str]) -> Iterator[str]:
    """Iterates over all files and dirs in 'sources' and returns
    only the filepaths that need fixing.
    """
    for source in sources:
        if os.path.isfile(source) and does_file_need_fix(source):
            yield source
        elif os.path.isdir(source):
            for root, _, filenames in os.walk(source):
                for filename in filenames:
                    filepath = os.path.join(root, filename)
                    if does_file_need_fix(filepath):
                        yield filepath


def does_file_need_fix(filepath: str) -> bool:
    if not re.search(r"\.pyi?$", filepath):
        return False
    with open(filepath, mode="r") as f:
        first_license_line = None
        for line in f:
            if line == license_header_lines[0]:
                first_license_line = line
                break
            elif line not in lines_to_keep:
                return True
        for header_line, line in zip(
            license_header_lines, chain((first_license_line,), f)
        ):
            if line != header_line:
                return True
    return False


def add_header_to_file(filepath: str) -> None:
    with open(filepath, mode="r") as f:
        lines = list(f)
    i = 0
    for i, line in enumerate(lines):
        if line not in lines_to_keep:
            break
    lines = lines[:i] + license_header_lines + lines[i:]
    with open(filepath, mode="w") as f:
        f.truncate()
        f.write("".join(lines))
    print(f"Fixed {os.path.relpath(filepath, os.getcwd())}")


def main():
    mode = sys.argv[1]
    assert mode in ("fix", "check")
    sources = [os.path.abspath(x) for x in sys.argv[2:]]
    files_to_fix = find_files_to_fix(sources)

    if mode == "fix":
        for filepath in files_to_fix:
            add_header_to_file(filepath)
    else:
        no_license_headers = list(files_to_fix)
        if no_license_headers:
            print("No license header found in:")
            cwd = os.getcwd()
            [
                print(f" - {os.path.relpath(filepath, cwd)}")
                for filepath in no_license_headers
            ]
            sys.exit(1)
        else:
            print("All files had license header")


if __name__ == "__main__":
    main()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from test_elasticsearch.test_cases import ElasticsearchTestCase


class TestIndices(ElasticsearchTestCase):
    def test_create_one_index(self):
        self.client.indices.create(index="test-index")
        self.assert_url_called("PUT", "/test-index")

    def test_delete_multiple_indices(self):
        self.client.indices.delete(index=["test-index", "second.index", "third/index"])
        self.assert_url_called("DELETE", "/test-index,second.index,third%2Findex")

    def test_exists_index(self):
        self.client.indices.exists(index="second.index,third/index")
        self.assert_url_called("HEAD", "/second.index,third%2Findex")

    def test_passing_empty_value_for_required_param_raises_exception(self):
        self.assertRaises(ValueError, self.client.indices.exists, index=None)
        self.assertRaises(ValueError, self.client.indices.exists, index=[])
        self.assertRaises(ValueError, self.client.indices.exists, index="")

    def test_put_mapping_without_index(self):
        self.client.indices.put_mapping(doc_type="doc-type", body={})
        self.assert_url_called("PUT", "/_all/doc-type/_mapping")



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import nox

SOURCE_FILES = (
    "setup.py",
    "noxfile.py",
    "elasticsearch/",
    "test_elasticsearch/",
    "utils/",
)


@nox.session(python=["2.7", "3.4", "3.5", "3.6", "3.7", "3.8", "3.9"])
def test(session):
    session.install(".")
    session.install("-r", "dev-requirements.txt")

    session.run("python", "setup.py", "test")


@nox.session()
def format(session):
    session.install("black", "isort")

    session.run("isort", "--profile=black", *SOURCE_FILES)
    session.run("black", "--target-version=py27", *SOURCE_FILES)
    session.run("python", "utils/license-headers.py", "fix", *SOURCE_FILES)

    lint(session)


@nox.session()
def lint(session):
    session.install("flake8", "black", "mypy", "isort", "types-requests")

    session.run("isort", "--check", "--profile=black", *SOURCE_FILES)
    session.run("black", "--target-version=py27", "--check", *SOURCE_FILES)
    session.run("flake8", *SOURCE_FILES)
    session.run("python", "utils/license-headers.py", "check", *SOURCE_FILES)

    # Workaround to make '-r' to still work despite uninstalling aiohttp below.
    session.run("python", "-m", "pip", "install", "aiohttp")

    # Run mypy on the package and then the type examples separately for
    # the two different mypy use-cases, ourselves and our users.
    session.run("mypy", "--strict", "elasticsearch/")
    session.run("mypy", "--strict", "test_elasticsearch/test_types/sync_types.py")
    session.run("mypy", "--strict", "test_elasticsearch/test_types/async_types.py")

    # Make sure we don't require aiohttp to be installed for users to
    # receive type hint information from mypy.
    session.run("python", "-m", "pip", "uninstall", "--yes", "aiohttp")
    session.run("mypy", "--strict", "elasticsearch/")
    session.run("mypy", "--strict", "test_elasticsearch/test_types/sync_types.py")


@nox.session()
def docs(session):
    session.install(".")
    session.install(
        "-rdev-requirements.txt", "sphinx-rtd-theme", "sphinx-autodoc-typehints"
    )
    session.run("python", "-m", "pip", "install", "sphinx-autodoc-typehints")

    session.run("sphinx-build", "docs/sphinx/", "docs/sphinx/_build", "-b", "html")



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

"""
Dynamically generated set of TestCases based on set of yaml files describing
some integration tests. These files are shared among all official Elasticsearch
clients.
"""
import io
import json
import os
import re
import sys
import warnings
import zipfile

import pytest
import urllib3
import yaml

from elasticsearch import ElasticsearchWarning, RequestError, TransportError
from elasticsearch.client.utils import _COMPATIBILITY_MIMETYPE, _base64_auth_header
from elasticsearch.compat import string_types
from elasticsearch.helpers.test import _get_version

from . import get_client

# some params had to be changed in python, keep track of them so we can rename
# those in the tests accordingly
PARAMS_RENAMES = {"type": "doc_type", "from": "from_"}
APIS_USING_TYPE_INSTEAD_OF_DOC_TYPE = {
    "nodes.hot_threads",
    "license.post_start_trial",
}

# mapping from catch values to http status codes
CATCH_CODES = {"missing": 404, "conflict": 409, "unauthorized": 401}

# test features we have implemented
IMPLEMENTED_FEATURES = {
    "gtelte",
    "stash_in_path",
    "headers",
    "catch_unauthorized",
    "default_shards",
    "warnings",
    "allowed_warnings",
    "allowed_warnings_regex",
    "contains",
    "arbitrary_key",
    "transform_and_set",
}

# broken YAML tests on some releases
SKIP_TESTS = {
    # Uses bad input intentionally
    "update/90_error[0]",
    "search/20_default_values[1]",
    # Warning about date_histogram.interval deprecation is raised randomly
    "search/aggregation/250_moving_fn[1]",
    # body: null
    "indices/simulate_index_template/10_basic[2]",
    # No ML node with sufficient capacity / random ML failing
    "ml/start_stop_datafeed",
    "ml/post_data",
    "ml/jobs_crud",
    "ml/datafeeds_crud",
    "ml/set_upgrade_mode",
    "ml/reset_job[2]",
    "ml/jobs_get_stats",
    "ml/get_datafeed_stats",
    "ml/get_trained_model_stats",
    "ml/delete_job_force",
    "ml/jobs_get_result_overall_buckets",
    "ml/bucket_correlation_agg[0]",
    "ml/job_groups",
    "transform/transforms_stats_continuous[0]",
    # Fails bad request instead of 404?
    "ml/inference_crud",
    # rollup/security_tests time out?
    "rollup/security_tests",
    # Our TLS certs are custom
    "ssl/10_basic[0]",
    # Our user is custom
    "users/10_basic[3]",
    # Shards/snapshots aren't right?
    "searchable_snapshots/10_usage[1]",
    # flaky data streams?
    "data_stream/10_basic[1]",
    "data_stream/80_resolve_index_data_streams[1]",
    # bad formatting?
    "cat/allocation/10_basic",
    # service account number not right?
    "service_accounts/10_basic[1]",
    # doesn't use 'contains' properly?
    "xpack/10_basic[0]",
    "privileges/40_get_user_privs[0]",
    "privileges/40_get_user_privs[1]",
    # bad use of 'is_false'?
    "indices/get_alias/10_basic[22]",
    # unique usage of 'set'
    "indices/stats/50_disk_usage[0]",
    "indices/stats/60_field_usage[0]",
}

APIS_WITH_BODY_FIELDS = {
    "search",
    "search_mvt",
    "scroll",
    "clear_scroll",
    "update",
    "indices.create",
}


XPACK_FEATURES = None
ES_VERSION = None
RUN_ASYNC_REST_API_TESTS = (
    sys.version_info >= (3, 6)
    and os.environ.get("PYTHON_CONNECTION_CLASS") == "RequestsHttpConnection"
)

FALSEY_VALUES = ("", None, False, 0, 0.0)

# Means the client will be emitting 'application/vnd.elasticsearch;compatible-with=X' headers
COMPATIBILITY_MODE_ENABLED = os.environ.get("ELASTIC_CLIENT_APIVERSIONING") in (
    "1",
    "true",
)
COMPATIBILITY_MIMETYPE = _COMPATIBILITY_MIMETYPE


class YamlRunner:
    def __init__(self, client):
        self.client = client
        self.last_response = None

        self._run_code = None
        self._setup_code = None
        self._teardown_code = None
        self._state = {}

    def use_spec(self, test_spec):
        self._setup_code = test_spec.pop("setup", None)
        self._run_code = test_spec.pop("run", None)
        self._teardown_code = test_spec.pop("teardown", None)

    def setup(self):
        # Pull skips from individual tests to not do unnecessary setup.
        skip_code = []
        for action in self._run_code:
            assert len(action) == 1
            action_type, _ = list(action.items())[0]
            if action_type == "skip":
                skip_code.append(action)
            else:
                break

        if self._setup_code or skip_code:
            self.section("setup")
        if skip_code:
            self.run_code(skip_code)
        if self._setup_code:
            self.run_code(self._setup_code)

    def teardown(self):
        if self._teardown_code:
            self.section("teardown")
            self.run_code(self._teardown_code)

    def es_version(self):
        global ES_VERSION
        if ES_VERSION is None:
            version_string = (self.client.info())["version"]["number"]
            if "." not in version_string:
                return ()
            version = version_string.strip().split(".")
            ES_VERSION = tuple(int(v) if v.isdigit() else 999 for v in version)
        return ES_VERSION

    def section(self, name):
        print(("=" * 10) + " " + name + " " + ("=" * 10))

    def run(self):
        try:
            self.setup()
            self.section("test")
            self.run_code(self._run_code)
        finally:
            try:
                self.teardown()
            except Exception:
                pass

    def run_code(self, test):
        """Execute an instruction based on it's type."""
        for action in test:
            assert len(action) == 1
            action_type, action = list(action.items())[0]
            print(action_type, action)

            if hasattr(self, "run_" + action_type):
                getattr(self, "run_" + action_type)(action)
            else:
                raise RuntimeError("Invalid action type %r" % (action_type,))

    def run_do(self, action):
        api = self.client
        headers = action.pop("headers", None)
        catch = action.pop("catch", None)
        warn = action.pop("warnings", ())
        allowed_warnings = action.pop("allowed_warnings", ())
        if isinstance(allowed_warnings, str):
            allowed_warnings = (allowed_warnings,)
        allowed_warnings_regex = action.pop("allowed_warnings_regex", ())
        if isinstance(allowed_warnings_regex, str):
            allowed_warnings_regex = (allowed_warnings_regex,)
        assert len(action) == 1

        # Remove the x_pack_rest_user authentication
        # if it's given via headers. We're already authenticated
        # via the 'elastic' user.
        if (
            headers
            and headers.get("Authorization", None)
            == "Basic eF9wYWNrX3Jlc3RfdXNlcjp4LXBhY2stdGVzdC1wYXNzd29yZA=="
        ):
            headers.pop("Authorization")

        if headers and "Content-Type" in headers and COMPATIBILITY_MODE_ENABLED:
            headers["Content-Type"] = COMPATIBILITY_MIMETYPE

        method, args = list(action.items())[0]
        args["headers"] = headers

        # locate api endpoint
        for m in method.split("."):
            # Some deprecated APIs are prefixed with 'xpack-*'
            if m.startswith("xpack-"):
                m = m.replace("xpack-", "")
            assert hasattr(api, m)
            api = getattr(api, m)

        # some parameters had to be renamed to not clash with python builtins,
        # compensate
        for k in PARAMS_RENAMES:

            # Don't do the 'doc_type' rename for APIs that actually want 'type'
            if k == "type" and method in APIS_USING_TYPE_INSTEAD_OF_DOC_TYPE:
                continue

            if k in args:
                args[PARAMS_RENAMES[k]] = args.pop(k)

        # resolve vars
        for k in args:
            args[k] = self._resolve(args[k])

        # If there's a body parameter given to an API with
        # body fields enabled we expand the body to parameters.
        if (
            "body" in args
            and isinstance(args["body"], dict)
            and method in APIS_WITH_BODY_FIELDS
        ):
            args.update(
                {PARAMS_RENAMES.get(k, k): v for k, v in args.pop("body").items()}
            )

        warnings.simplefilter("always", category=ElasticsearchWarning)
        with warnings.catch_warnings(record=True) as caught_warnings:
            try:
                self.last_response = api(**args)
            except Exception as e:
                if not catch:
                    raise
                self.run_catch(catch, e)
            else:
                if catch:
                    raise AssertionError(
                        "Failed to catch %r in %r." % (catch, self.last_response)
                    )

        # Filter out warnings raised by other components.
        caught_warnings = [
            str(w.message)
            for w in caught_warnings
            if w.category == ElasticsearchWarning
            and (not allowed_warnings or str(w.message) not in allowed_warnings)
            and (
                not allowed_warnings_regex
                or all(
                    re.search(pattern, str(w.message)) is None
                    for pattern in allowed_warnings_regex
                )
            )
        ]

        # This warning can show up in many places but isn't accounted for
        # in tests, so we remove it to make sure things pass.
        include_type_name_warning = (
            "[types removal] Using include_type_name in create index requests is deprecated. "
            "The parameter will be removed in the next major version."
        )
        if (
            include_type_name_warning in caught_warnings
            and include_type_name_warning not in warn
        ):
            caught_warnings.remove(include_type_name_warning)

        # Sorting removes the issue with order raised. We only care about
        # if all warnings are raised in the single API call.
        if warn and sorted(warn) != sorted(caught_warnings):
            raise AssertionError(
                "Expected warnings not equal to actual warnings: expected=%r actual=%r"
                % (warn, caught_warnings)
            )

    def run_catch(self, catch, exception):
        if catch == "param":
            assert isinstance(exception, TypeError)
            return

        assert isinstance(exception, TransportError)
        if catch in CATCH_CODES:
            assert CATCH_CODES[catch] == exception.status_code
        elif catch[0] == "/" and catch[-1] == "/":
            assert (
                re.search(catch[1:-1], exception.error + " " + repr(exception.info)),
                "%s not in %r" % (catch, exception.info),
            ) is not None
        self.last_response = exception.info

    def run_skip(self, skip):
        global IMPLEMENTED_FEATURES

        if "features" in skip:
            features = skip["features"]
            if not isinstance(features, (tuple, list)):
                features = [features]
            for feature in features:
                if feature in IMPLEMENTED_FEATURES:
                    continue
                pytest.skip("feature '%s' is not supported" % feature)

        if "version" in skip:
            version, reason = skip["version"], skip["reason"]
            if version == "all":
                pytest.skip(reason)
            min_version, max_version = version.split("-")
            min_version = _get_version(min_version) or (0,)
            max_version = _get_version(max_version) or (999,)
            if min_version <= (self.es_version()) <= max_version:
                pytest.skip(reason)

    def run_gt(self, action):
        for key, value in action.items():
            value = self._resolve(value)
            assert self._lookup(key) > value

    def run_gte(self, action):
        for key, value in action.items():
            value = self._resolve(value)
            assert self._lookup(key) >= value

    def run_lt(self, action):
        for key, value in action.items():
            value = self._resolve(value)
            assert self._lookup(key) < value

    def run_lte(self, action):
        for key, value in action.items():
            value = self._resolve(value)
            assert self._lookup(key) <= value

    def run_set(self, action):
        for key, value in action.items():
            value = self._resolve(value)
            self._state[value] = self._lookup(key)

    def run_is_false(self, action):
        try:
            value = self._lookup(action)
        except AssertionError:
            pass
        else:
            assert value in FALSEY_VALUES

    def run_is_true(self, action):
        value = self._lookup(action)
        assert value not in FALSEY_VALUES

    def run_length(self, action):
        for path, expected in action.items():
            value = self._lookup(path)
            expected = self._resolve(expected)
            assert expected == len(value)

    def run_match(self, action):
        for path, expected in action.items():
            value = self._lookup(path)
            expected = self._resolve(expected)

            if (
                isinstance(expected, string_types)
                and expected.startswith("/")
                and expected.endswith("/")
            ):
                expected = re.compile(expected[1:-1], re.VERBOSE | re.MULTILINE)
                assert expected.search(value), "%r does not match %r" % (
                    value,
                    expected,
                )
            elif isinstance(value, list) and isinstance(expected, list):
                assert len(value) == len(
                    expected
                ), "Length between %r and %r wasn't equal" % (value, expected)
                [self._assert_match_equals(a, b) for a, b in zip(value, expected)]
            else:
                self._assert_match_equals(value, expected)

    def run_contains(self, action):
        for path, expected in action.items():
            value = self._lookup(path)  # list[dict[str,str]] is returned
            expected = self._resolve(expected)  # dict[str, str]

            if expected not in value:
                raise AssertionError("%s is not contained by %s" % (expected, value))

    def run_transform_and_set(self, action):
        for key, value in action.items():
            # Convert #base64EncodeCredentials(id,api_key) to ["id", "api_key"]
            if "#base64EncodeCredentials" in value:
                value = value.replace("#base64EncodeCredentials", "")
                value = value.replace("(", "").replace(")", "").split(",")
                self._state[key] = _base64_auth_header(
                    (self._lookup(value[0]), self._lookup(value[1]))
                )

    def _resolve(self, value):
        # resolve variables
        if isinstance(value, string_types) and "$" in value:
            for k, v in self._state.items():
                for key_replace in ("${" + k + "}", "$" + k):
                    if value == key_replace:
                        value = v
                        break
                    # We only do the in-string replacement if using ${...}
                    elif (
                        key_replace.startswith("${")
                        and isinstance(value, string_types)
                        and key_replace in value
                    ):
                        value = value.replace(key_replace, str(v))
                        break
                    # We only do the in-string replacement if value is JSON string
                    # E.g. '{\n  "password_hash" : "$hash"\n}\n'
                    elif (
                        key_replace.startswith("$")
                        and isinstance(value, string_types)
                        and key_replace in value
                        and not value.startswith("$")
                    ):
                        value = value.replace(key_replace, str(v))
                        break

        if isinstance(value, string_types):
            value = value.strip()
        elif isinstance(value, dict):
            value = dict((k, self._resolve(v)) for (k, v) in value.items())
        elif isinstance(value, list):
            value = list(map(self._resolve, value))
        return value

    def _lookup(self, path):
        # fetch the possibly nested value from last_response
        value = self.last_response
        if path == "$body":
            return value
        path = path.replace(r"\.", "\1")
        for step in path.split("."):
            if not step:
                continue
            step = step.replace("\1", ".")
            step = self._resolve(step)

            if (
                isinstance(step, string_types)
                and step.isdigit()
                and isinstance(value, list)
            ):
                step = int(step)
                assert isinstance(value, list)
                assert len(value) > step
            elif step == "_arbitrary_key_":
                return list(value.keys())[0]
            else:
                assert step in value
            value = value[step]
        return value

    def _feature_enabled(self, name):
        global XPACK_FEATURES, IMPLEMENTED_FEATURES
        if XPACK_FEATURES is None:
            try:
                xinfo = self.client.xpack.info()
                XPACK_FEATURES = set(
                    f for f in xinfo["features"] if xinfo["features"][f]["enabled"]
                )
                IMPLEMENTED_FEATURES.add("xpack")
            except RequestError:
                XPACK_FEATURES = set()
                IMPLEMENTED_FEATURES.add("no_xpack")
        return name in XPACK_FEATURES

    def _assert_match_equals(self, a, b):
        # Handle for large floating points with 'E'
        if isinstance(b, string_types) and isinstance(a, float) and "e" in repr(a):
            a = repr(a).replace("e+", "E")

        assert a == b, "%r does not match %r" % (a, b)


@pytest.fixture(scope="function")
def sync_runner(sync_client):
    return YamlRunner(sync_client)


YAML_TEST_SPECS = []

# Try loading the REST API test specs from the Elastic Artifacts API
try:
    # Construct the HTTP and Elasticsearch client
    http = urllib3.PoolManager(
        retries=10, headers=urllib3.util.make_headers(accept_encoding=True)
    )
    client = get_client()

    # If we're running in compatibility mode the server won't have the previous versions'
    # test suite so we use the overridden 'STACK_VERSION' in run-repository.sh instead.
    if os.environ.get("STACK_VERSION") and COMPATIBILITY_MODE_ENABLED:
        version_number = os.environ["STACK_VERSION"]
        # Setting 'build_hash' to empty means we'll always get the latest build.
        build_hash = ""
    else:
        # Make a request to Elasticsearch for the build hash, we'll be looking for
        # an artifact with this same hash to download test specs for.
        client_info = client.info()
        version_number = client_info["version"]["number"]
        build_hash = client_info["version"]["build_hash"]

    # Now talk to the artifacts API with the 'STACK_VERSION' environment variable
    resp = http.request(
        "GET",
        "https://artifacts-api.elastic.co/v1/versions/%s" % (version_number,),
    )
    resp = json.loads(resp.data.decode("utf-8"))

    # Look through every build and see if one matches the commit hash
    # we're looking for. If not it's okay, we'll just use the latest and
    # hope for the best!
    builds = resp["version"]["builds"]
    for build in builds:
        if build["projects"]["elasticsearch"]["commit_hash"] == build_hash:
            break
    else:
        build = builds[0]  # Use the latest

    # Now we're looking for the 'rest-api-spec-<VERSION>-sources.jar' file
    # to download and extract in-memory.
    packages = build["projects"]["elasticsearch"]["packages"]
    for package in packages:
        if re.match(r"rest-resources-zip-.*\.zip", package):
            package_url = packages[package]["url"]
            break
    else:
        raise RuntimeError(
            "Could not find the package 'rest-resources-zip-*.zip' in build %r" % build
        )

    # Download the zip and start reading YAML from the files in memory
    package_zip = zipfile.ZipFile(io.BytesIO(http.request("GET", package_url).data))
    for yaml_file in package_zip.namelist():
        yaml_filter_pattern = r"^rest-api-spec/%s/.*\.ya?ml$" % (
            "compatTest" if COMPATIBILITY_MODE_ENABLED else "test"
        )
        if not re.match(yaml_filter_pattern, yaml_file):
            continue
        yaml_tests = list(yaml.safe_load_all(package_zip.read(yaml_file)))

        # Each file may have a "test" named 'setup' or 'teardown',
        # these sets of steps should be run at the beginning and end
        # of every other test within the file so we do one pass to capture those.
        setup_steps = teardown_steps = None
        test_numbers_and_steps = []
        test_number = 0

        for yaml_test in yaml_tests:
            test_name, test_step = yaml_test.popitem()
            if test_name == "setup":
                setup_steps = test_step
            elif test_name == "teardown":
                teardown_steps = test_step
            else:
                test_numbers_and_steps.append((test_number, test_step))
                test_number += 1

        # Now we combine setup, teardown, and test_steps into
        # a set of pytest.param() instances
        for test_number, test_step in test_numbers_and_steps:
            # Build the id from the name of the YAML file and
            # the number within that file. Most important step
            # is to remove most of the file path prefixes and
            # the .yml suffix.
            pytest_test_name = yaml_file.rpartition(".")[0].replace(".", "/")
            for prefix in (
                "rest-api-spec/",
                "compatTest/",
                "test/",
                "free/",
                "platinum/",
            ):
                if pytest_test_name.startswith(prefix):
                    pytest_test_name = pytest_test_name[len(prefix) :]
            pytest_param_id = "%s[%d]" % (pytest_test_name, test_number)

            pytest_param = {
                "setup": setup_steps,
                "run": test_step,
                "teardown": teardown_steps,
            }
            # Skip either 'test_name' or 'test_name[x]'
            if pytest_test_name in SKIP_TESTS or pytest_param_id in SKIP_TESTS:
                pytest_param["skip"] = True

            YAML_TEST_SPECS.append(pytest.param(pytest_param, id=pytest_param_id))

except Exception as e:
    warnings.warn("Could not load REST API tests: %s" % (str(e),))

# Sort the tests by ID so they're grouped together nicely.
YAML_TEST_SPECS = sorted(YAML_TEST_SPECS, key=lambda param: param.id)


if not RUN_ASYNC_REST_API_TESTS:

    @pytest.mark.parametrize("test_spec", YAML_TEST_SPECS)
    def test_rest_api_spec(test_spec, sync_runner):
        if test_spec.get("skip", False):
            pytest.skip("Manually skipped in 'SKIP_TESTS'")
        sync_runner.use_spec(test_spec)
        sync_runner.run()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from collections import defaultdict
from unittest import SkipTest  # noqa: F401
from unittest import TestCase

from elasticsearch import Elasticsearch


class DummyTransport(object):
    def __init__(self, hosts, responses=None, **kwargs):
        self.hosts = hosts
        self.responses = responses
        self.call_count = 0
        self.calls = defaultdict(list)

    def perform_request(self, method, url, params=None, headers=None, body=None):
        resp = 200, {}
        if self.responses:
            resp = self.responses[self.call_count]
        self.call_count += 1
        self.calls[(method, url)].append((params, headers, body))
        return resp


class ElasticsearchTestCase(TestCase):
    def setUp(self):
        super(ElasticsearchTestCase, self).setUp()
        self.client = Elasticsearch(transport_class=DummyTransport)

    def assert_call_count_equals(self, count):
        self.assertEqual(count, self.client.transport.call_count)

    def assert_url_called(self, method, url, count=1):
        self.assertIn((method, url), self.client.transport.calls)
        calls = self.client.transport.calls[(method, url)]
        self.assertEqual(count, len(calls))
        return calls


class TestElasticsearchTestCase(ElasticsearchTestCase):
    def test_our_transport_used(self):
        self.assertIsInstance(self.client.transport, DummyTransport)

    def test_start_with_0_call(self):
        self.assert_call_count_equals(0)

    def test_each_call_is_recorded(self):
        self.client.transport.perform_request("GET", "/")
        self.client.transport.perform_request("DELETE", "/42", params={}, body="body")
        self.assert_call_count_equals(2)
        self.assertEqual(
            [({}, None, "body")], self.assert_url_called("DELETE", "/42", 1)
        )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import re
from unittest import SkipTest

import pytest
from mock import patch

from elasticsearch import (
    Elasticsearch,
    RequestError,
    RequestsHttpConnection,
    Urllib3HttpConnection,
)
from elasticsearch.helpers.test import CA_CERTS, ELASTICSEARCH_URL


@pytest.fixture(scope="function")
def mvt_setup(sync_client):
    sync_client.indices.create(
        index="museums",
        body={
            "mappings": {
                "properties": {
                    "location": {"type": "geo_point"},
                    "name": {"type": "keyword"},
                    "price": {"type": "long"},
                    "included": {"type": "boolean"},
                }
            }
        },
    )
    sync_client.bulk(
        index="museums",
        body=[
            {"index": {"_id": "1"}},
            {
                "location": "52.374081,4.912350",
                "name": "NEMO Science Museum",
                "price": 1750,
                "included": True,
            },
            {"index": {"_id": "2"}},
            {
                "location": "52.369219,4.901618",
                "name": "Museum Het Rembrandthuis",
                "price": 1500,
                "included": False,
            },
            {"index": {"_id": "3"}},
            {
                "location": "52.371667,4.914722",
                "name": "Nederlands Scheepvaartmuseum",
                "price": 1650,
                "included": True,
            },
            {"index": {"_id": "4"}},
            {
                "location": "52.371667,4.914722",
                "name": "Amsterdam Centre for Architecture",
                "price": 0,
                "included": True,
            },
        ],
        refresh=True,
    )


@pytest.mark.parametrize(
    "connection_class", [Urllib3HttpConnection, RequestsHttpConnection]
)
def test_mapbox_vector_tile_logging(mvt_setup, connection_class):
    client = Elasticsearch(
        ELASTICSEARCH_URL, connection_class=connection_class, ca_certs=CA_CERTS
    )
    client.info()

    with patch("elasticsearch.connection.base.logger") as logger:
        client.search_mvt(
            index="museums",
            zoom=13,
            x=4207,
            y=2692,
            field="location",
        )

    assert logger.info.call_count == 1
    assert re.search(
        r"^POST https?://[^/]+/museums/_mvt/location/13/4207/2692 \[status:200 request:0\.[0-9]{3}s\]$",
        logger.info.call_args_list[0][0][0] % logger.info.call_args_list[0][0][1:],
    )

    assert logger.debug.call_count == 2
    req, resp = logger.debug.call_args_list
    assert req[0] == ("> %s", None)
    assert re.search(
        r"< b'.+'$",
        resp[0][0] % (resp[0][1:]),
        flags=re.DOTALL,
    )

    # Errors should still be JSON
    with patch("elasticsearch.connection.base.logger") as logger, pytest.raises(
        RequestError
    ) as e:
        client.search_mvt(
            index="museums",
            zoom=-100,
            x=4207,
            y=2692,
            field="location",
        )

    assert e.value.info == {
        "error": {
            "root_cause": [
                {
                    "type": "illegal_argument_exception",
                    "reason": "Invalid geotile_grid precision of -100. Must be between 0 and 29.",
                }
            ],
            "type": "illegal_argument_exception",
            "reason": "Invalid geotile_grid precision of -100. Must be between 0 and 29.",
        },
        "status": 400,
    }
    assert e.value.status_code == 400

    assert logger.warning.call_count == 1
    assert re.search(
        r"^POST https?://[^/]+/museums/_mvt/location/-100/4207/2692 \[status:400 request:0\.[0-9]{3}s\]$",
        logger.warning.call_args_list[0][0][0]
        % logger.warning.call_args_list[0][0][1:],
    )

    assert logger.debug.call_count == 2
    req, resp = logger.debug.call_args_list
    assert req[0] == ("> %s", None)

    # The JSON error body is still logged properly.
    assert resp[0][0] % (resp[0][1:]) == (
        '< {"error":{"root_cause":[{"type":"illegal_argument_exception","reason":"Invalid '
        'geotile_grid precision of -100. Must be between 0 and 29."}],"type":"illegal_argument_exception",'
        '"reason":"Invalid geotile_grid precision of -100. Must be between 0 and 29."},"status":400}'
    )


@pytest.mark.parametrize(
    "connection_class", [Urllib3HttpConnection, RequestsHttpConnection]
)
def test_mapbox_vector_tile_response(mvt_setup, connection_class):
    client = Elasticsearch(
        ELASTICSEARCH_URL, connection_class=connection_class, ca_certs=CA_CERTS
    )

    resp = client.search_mvt(
        index="museums",
        zoom=13,
        x=4207,
        y=2692,
        field="location",
        grid_precision=2,
        fields=["name", "price"],
        query={"term": {"included": True}},
        aggs={
            "min_price": {"min": {"field": "price"}},
            "max_price": {"max": {"field": "price"}},
            "avg_price": {"avg": {"field": "price"}},
        },
    )
    assert isinstance(resp, bytes)

    try:
        import mapbox_vector_tile
    except ImportError:
        raise SkipTest("Requires the 'mapbox-vector-tile' package")

    # Decode the binary as MVT
    tile = mapbox_vector_tile.decode(resp)

    # Assert some general things about the structure, mostly we want
    # to know that we got back a valid MVT.
    assert set(tile.keys()) == {"hits", "aggs", "meta"}
